/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

//{{NO_DEPENDENCIES}}
// Used by ClangResource.rc
//
#define IDR_HEADER_COMPILE_TIME 101
#define IDR_HEADER_PRE_COMPILE 102

#define IDR_PCH_OPENCL_32 110
#define IDR_PCH_OPENCL_64 111

///////////////////////////////////////////

#define IDR_BC_START 120
#define IDR_BC_COMMON 120
#define IDR_BC_SIZE_T_32 124
#define IDR_BC_SIZE_T_64 125
#define IDR_BC_RS_COMMON 126
#define IDR_BC_RS_V1 127
#define IDR_BC_RS_V2 128
#define IDR_BC_SPIRV_COMMON 129
#define IDR_BC_SPIRV_SIZE_T_32 130
#define IDR_BC_SPIRV_SIZE_T_64 131
#define IDR_BC_END 132
#define IDR_CTH_H 133

///////////////////////////////////////////

// Next default values for new objects
//
#ifdef APSTUDIO_INVOKED
#ifndef APSTUDIO_READONLY_SYMBOLS
#define _APS_NEXT_RESOURCE_VALUE 140
#define _APS_NEXT_COMMAND_VALUE 40001
#define _APS_NEXT_CONTROL_VALUE 1001
#define _APS_NEXT_SYMED_VALUE 141
#endif
#endif
