/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

DECLARE_CISA_OPCODE(EOPCODE_INVALID, "invalid", ISA_RESERVED_0)
DECLARE_CISA_OPCODE(EOPCODE_MOV, "mov", ISA_MOV)
DECLARE_CISA_OPCODE(EOPCODE_ADD, "add", ISA_ADD)
DECLARE_CISA_OPCODE(EOPCODE_MUL, "mul", ISA_MUL)
DECLARE_CISA_OPCODE(EOPCODE_POW, "pow", ISA_POW)
DECLARE_CISA_OPCODE(EOPCODE_DIV, "div", ISA_DIV)
DECLARE_CISA_OPCODE(EOPCODE_AVG, "avg", ISA_AVG)
DECLARE_CISA_OPCODE(EOPCODE_PLN, "plane", ISA_PLANE)
DECLARE_CISA_OPCODE(EOPCODE_CMP, "cmp", ISA_CMP)
DECLARE_CISA_OPCODE(EOPCODE_SEL, "sel", ISA_SEL)
DECLARE_CISA_OPCODE(EOPCODE_COS, "cos", ISA_COS)
DECLARE_CISA_OPCODE(EOPCODE_SIN, "sin", ISA_SIN)
DECLARE_CISA_OPCODE(EOPCODE_LOG, "log", ISA_LOG)
DECLARE_CISA_OPCODE(EOPCODE_EXP, "exp", ISA_EXP)
DECLARE_CISA_OPCODE(EOPCODE_FRC, "frc", ISA_FRC)
DECLARE_CISA_OPCODE(EOPCODE_SQRT, "sqrt", ISA_SQRT)
DECLARE_CISA_OPCODE(EOPCODE_RNDD, "rndd", ISA_RNDD)
DECLARE_CISA_OPCODE(EOPCODE_RNDU, "rndu", ISA_RNDU)
DECLARE_CISA_OPCODE(EOPCODE_RNDZ, "rndz", ISA_RNDZ)
DECLARE_CISA_OPCODE(EOPCODE_RNDE, "rnde", ISA_RNDE)
DECLARE_CISA_OPCODE(EOPCODE_RSQRT, "rsqrt", ISA_RSQRT)
DECLARE_CISA_OPCODE(EOPCODE_INV, "inv", ISA_INV)
DECLARE_CISA_OPCODE(EOPCODE_MAD, "mad", ISA_MAD)
DECLARE_CISA_OPCODE(EOPCODE_LRP, "lrp", ISA_LRP)
DECLARE_CISA_OPCODE(EOPCODE_SEND, "send", ISA_RAW_SEND)
DECLARE_CISA_OPCODE(EOPCODE_GOTO, "goto", ISA_GOTO)
DECLARE_CISA_OPCODE(EOPCODE_LABEL, "label", ISA_LABEL)
DECLARE_CISA_OPCODE(EOPCODE_XOR, "xor", ISA_XOR)
DECLARE_CISA_OPCODE(EOPCODE_OR, "or", ISA_OR)
DECLARE_CISA_OPCODE(EOPCODE_AND, "and", ISA_AND)
DECLARE_CISA_OPCODE(EOPCODE_NOT, "not", ISA_NOT)
DECLARE_CISA_OPCODE(EOPCODE_SETP, "setp", ISA_SETP)
DECLARE_CISA_OPCODE(EOPCODE_SHL, "shl", ISA_SHL)
DECLARE_CISA_OPCODE(EOPCODE_SHR, "shr", ISA_SHR)
DECLARE_CISA_OPCODE(EOPCODE_ROL, "rol", ISA_ROL)
DECLARE_CISA_OPCODE(EOPCODE_ROR, "ror", ISA_ROR)
DECLARE_CISA_OPCODE(EOPCODE_ADD3, "add3", ISA_ADD3)
DECLARE_CISA_OPCODE(EOPCODE_BFN, "bfn", ISA_BFN)
DECLARE_CISA_OPCODE(EOPCODE_OWLOAD, "oword_ld_unaligned", ISA_OWORD_LD_UNALIGNED)
DECLARE_CISA_OPCODE(EOPCODE_GATHER, "gather", ISA_GATHER)
DECLARE_CISA_OPCODE(EOPCODE_TYPED4_READ, "typed4_read", ISA_GATHER4_TYPED)
DECLARE_CISA_OPCODE(EOPCODE_SCATTER, "scatter", ISA_SCATTER)
DECLARE_CISA_OPCODE(EOPCODE_TYPED4_WRITE, "typed4_write", ISA_SCATTER4_TYPED)
DECLARE_CISA_OPCODE(EOPCODE_MULH, "mulh", ISA_MULH)
DECLARE_CISA_OPCODE(EOPCODE_MOD, "mod", ISA_MOD)
DECLARE_CISA_OPCODE(EOPCODE_ASR, "asr", ISA_ASR)
DECLARE_CISA_OPCODE(EOPCODE_MIN, "min", ISA_FMINMAX)
DECLARE_CISA_OPCODE(EOPCODE_MAX, "max", ISA_FMINMAX)
DECLARE_CISA_OPCODE(EOPCODE_UADDC, "uaddc", ISA_ADDC)
DECLARE_CISA_OPCODE(EOPCODE_USUBB, "usubb", ISA_SUBB)
DECLARE_CISA_OPCODE(EOPCODE_BARRIER, "barrier", ISA_BARRIER)
DECLARE_CISA_OPCODE(EOPCODE_BFI, "bfi", ISA_BFI)
DECLARE_CISA_OPCODE(EOPCODE_BFE, "bfe", ISA_BFE)
DECLARE_CISA_OPCODE(EOPCODE_BFREV, "bfrev", ISA_BFREV)
