/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2026 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

#pragma once

#include "LLVMWarningsPush.hpp"
#include "llvm/Config/llvm-config.h"
#include "llvm/PassRegistry.h"
#include "LLVMWarningsPop.hpp"

namespace llvm {
class PassRegistry;
}

// The following declarations are placed according to alphabetic order for simplicity
void initializeAccuracyDecoratedCallsBiFResolutionLPMPass(llvm::PassRegistry &);
void initializeAddImplicitArgsLPMPass(llvm::PassRegistry &);
void initializeAddressSpaceAliasAnalysisPass(llvm::PassRegistry &);
void initializeAnnotateUniformAllocasPass(llvm::PassRegistry &);
void initializeAtomicOptPassPass(llvm::PassRegistry &);
void initializeAggregateArgumentsAnalysisLPMPass(llvm::PassRegistry &);
void initializeAlignmentAnalysisLPMPass(llvm::PassRegistry &);
void initializePreBIImportAnalysisLPMPass(llvm::PassRegistry &);
void initializePromoteToPredicatedMemoryAccessPass(llvm::PassRegistry &);
void initializeBIImportLPMPass(llvm::PassRegistry &);
void initializeBlockCoalescingPass(llvm::PassRegistry &);
void initializeBlockMemOpAddrScalarizationPassPass(llvm::PassRegistry &);
void initializeBreakConstantExprLPMPass(llvm::PassRegistry &);
void initializeBuiltinCallGraphAnalysisLPMPass(llvm::PassRegistry &);
void initializeBuiltinsConverterLPMPass(llvm::PassRegistry &);
void initializeCanonicalizeMulAddPass(llvm::PassRegistry &);
void initializeCastToGASAnalysisPass(llvm::PassRegistry &);
void initializeCastToGASInfoPass(llvm::PassRegistry &);
void initializeCastToGASInfoWrapperPass(llvm::PassRegistry &);
void initializeClearTessFactorsPass(llvm::PassRegistry &);
void initializeCleanPHINodePass(llvm::PassRegistry &);
void initializeCoalescingEnginePass(llvm::PassRegistry &);
void initializeCodeGenContextWrapperPass(llvm::PassRegistry &);
void initializeCodeGenPatternMatchPass(llvm::PassRegistry &);
void initializeConstantCoalescingPass(llvm::PassRegistry &);
void initializeCorrectlyRoundedDivSqrtLPMPass(llvm::PassRegistry &);
void initializeCrossPhaseConstPropPass(llvm::PassRegistry &);
void initializeCustomLoopVersioningPass(llvm::PassRegistry &);
void initializeCustomSafeOptPassPass(llvm::PassRegistry &);
void initializeCustomUnsafeOptPassPass(llvm::PassRegistry &);
void initializeHoistConvOpToDomPass(llvm::PassRegistry &);
void initializeHoistFMulInLoopPassPass(llvm::PassRegistry &);
void initializeHandleFRemInstructionsLPMPass(llvm::PassRegistry &);
void initializeDeSSAPass(llvm::PassRegistry &);
void initializeDetectCSWalkOrderPass(llvm::PassRegistry &);
void initializeSelectCSWalkOrderPass(llvm::PassRegistry &);
void initializeDeviceEnqueueFuncsAnalysisLPMPass(llvm::PassRegistry &);
void initializeDeviceEnqueueFuncsResolutionLPMPass(llvm::PassRegistry &);
void initializeDisableLoopUnrollOnRetryPass(llvm::PassRegistry &);
void initializeDpasScanPass(llvm::PassRegistry &);
void initializeDynamicTextureFoldingPass(llvm::PassRegistry &);
void initializeEarlyOutPatternsPass(llvm::PassRegistry &);
void initializeExtensionArgAnalysisLPMPass(llvm::PassRegistry &);
void initializeExtensionFuncsAnalysisLPMPass(llvm::PassRegistry &);
void initializeExtensionFuncsResolutionLPMPass(llvm::PassRegistry &);
void initializeErrorCheckLPMPass(llvm::PassRegistry &);
void initializePoisonFP64KernelsLPMPass(llvm::PassRegistry &);
void initializeReduceOptPassLPMPass(llvm::PassRegistry &);
void initializeTypesLegalizationPassLPMPass(llvm::PassRegistry &);
void initializeGenericCastToPtrOptLPMPass(llvm::PassRegistry &);
void initializeGASResolvingLPMPass(llvm::PassRegistry &);
void initializeGASRetValuePropagatorLPMPass(llvm::PassRegistry &);
void initializeStaticGASResolutionLPMPass(llvm::PassRegistry &);
void initializeLowerGPCallArgLPMPass(llvm::PassRegistry &);
void initializeGenerateBlockMemOpsPassPass(llvm::PassRegistry &);
void initializeGenericAddressAnalysisPass(llvm::PassRegistry &);
void initializeGenericAddressDynamicResolutionPass(llvm::PassRegistry &);
void initializeGenericNullPtrPropagationPass(llvm::PassRegistry &);
void initializeGenFDIVEmulationPass(llvm::PassRegistry &);
void initializeGenIRLoweringPass(llvm::PassRegistry &);
void initializeGEPLoopStrengthReductionPass(llvm::PassRegistry &);
void initializeGEPLoweringPass(llvm::PassRegistry &);
void initializeGenSpecificPatternPass(llvm::PassRegistry &);
void initializeGreedyLiveRangeReductionPass(llvm::PassRegistry &);
void initializeClampICBOOBAccessPass(llvm::PassRegistry &);
void initializeIGCIndirectICBPropagaionPass(llvm::PassRegistry &);
void initializeGenUpdateCBPass(llvm::PassRegistry &);
void initializeGenStrengthReductionPass(llvm::PassRegistry &);
void initializeGenOptLegalizerPass(llvm::PassRegistry &);
void initializeNanHandlingPass(llvm::PassRegistry &);
void initializeFixResourcePtrLPMPass(llvm::PassRegistry &);
void initializeFlattenSmallSwitchPass(llvm::PassRegistry &);
void initializeSplitIndirectEEtoSelPass(llvm::PassRegistry &);
void initializeExpandNonUniformInsertElementPass(llvm::PassRegistry &);
void initializePropagateCmpUniformityPass(llvm::PassRegistry &);
void initializeGenXFunctionGroupAnalysisPass(llvm::PassRegistry &);
void initializeGenXCodeGenModulePass(llvm::PassRegistry &);
void initializeEstimateFunctionSizePass(llvm::PassRegistry &);
void initializeSubroutineInlinerPass(llvm::PassRegistry &);
void initializeHandleLoadStoreInstructionsPass(llvm::PassRegistry &);
void initializeIGCConstPropPass(llvm::PassRegistry &);
void initializeJointMatrixFuncsResolutionPassLPMPass(llvm::PassRegistry &);
void initializeGatingSimilarSamplesPass(llvm::PassRegistry &);
void initializeImageFuncResolutionLPMPass(llvm::PassRegistry &);
void initializeImageFuncsAnalysisLPMPass(llvm::PassRegistry &);
void initializeImage3dToImage2darrayLPMPass(llvm::PassRegistry &);
void initializeInlineLocalsResolutionLPMPass(llvm::PassRegistry &);
void initializeInlineUnmaskedFunctionsPassPass(llvm::PassRegistry &);
void initializeInsertBranchOptPass(llvm::PassRegistry &);
void initializeHFOptPass(llvm::PassRegistry &);
void initializeDropTargetBBsPass(llvm::PassRegistry &);
void initializeLayoutPass(llvm::PassRegistry &);
void initializeLdShrinkPass(llvm::PassRegistry &);
void initializeLegalizationPass(llvm::PassRegistry &);
void initializeLegalizeResourcePointerPass(llvm::PassRegistry &);
void initializeLegalizeFunctionSignaturesLPMPass(llvm::PassRegistry &);
void initializeLiveVarsAnalysisPass(llvm::PassRegistry &);
void initializeLogicalAndToBranchPass(llvm::PassRegistry &);
void initializeLowerByValAttributePass(llvm::PassRegistry &);
void initializeLowerGEPForPrivMemPass(llvm::PassRegistry &);
void initializeLowerImplicitArgIntrinsicsPass(llvm::PassRegistry &);
void initializeLowerSIMDSizePass(llvm::PassRegistry &);
void initializeLowPrecisionOptPass(llvm::PassRegistry &);
void initializeMarkReadOnlyLoadPass(llvm::PassRegistry &);
void initializeMetaDataUtilsWrapperInitializerPass(llvm::PassRegistry &);
void initializeMetaDataUtilsWrapperPass(llvm::PassRegistry &);
void initializeMatchCommonKernelPatternsPass(llvm::PassRegistry &);
void initializeOpenCLPrintfAnalysisLPMPass(llvm::PassRegistry &);
void initializeOpenCLPrintfResolutionLPMPass(llvm::PassRegistry &);
void initializePeepholeTypeLegalizerPass(llvm::PassRegistry &);
void initializePositionDepAnalysisPass(llvm::PassRegistry &);
void initializePrepareInlineSamplerForBindlessLPMPass(llvm::PassRegistry &);
void initializePrintFPRangeAnalysisPass(llvm::PassRegistry &);
void initializePrivateMemoryResolutionPass(llvm::PassRegistry &);
void initializePrivateMemoryToSLMPass(llvm::PassRegistry &);
void initializePrivateMemoryUsageAnalysisLPMPass(llvm::PassRegistry &);
void initializeProcessFuncAttributesLPMPass(llvm::PassRegistry &);
void initializeProcessBuiltinMetaDataLPMPass(llvm::PassRegistry &);
void initializeInsertDummyKernelForSymbolTableLPMPass(llvm::PassRegistry &);
void initializeCleanupIndirectlyReferencedFunctionsLPMPass(llvm::PassRegistry &);
void initializeProgramScopeConstantAnalysisLPMPass(llvm::PassRegistry &);
void initializeProgramScopeConstantResolutionLPMPass(llvm::PassRegistry &);
void initializePromoteResourceToDirectASPass(llvm::PassRegistry &);
void initializePromoteStatelessToBindlessPass(llvm::PassRegistry &);
void initializePruneUnusedArgumentsPass(llvm::PassRegistry &);
void initializePullConstantHeuristicsPass(llvm::PassRegistry &);
void initializeScalarizerCodeGenPass(llvm::PassRegistry &);
void initializeReduceLocalPointersPass(llvm::PassRegistry &);
void initializeReplaceUnsupportedIntrinsicsPass(llvm::PassRegistry &);
void initializeResolveImageImplicitArgsForBindlessLPMPass(llvm::PassRegistry &);
void initializeResolveInlineSamplerForBindlessLPMPass(llvm::PassRegistry &);
void initializePreCompiledFuncImportPass(llvm::PassRegistry &);
void initializePurgeMetaDataUtilsLPMPass(llvm::PassRegistry &);
void initializeRemoveNonPositionOutputPass(llvm::PassRegistry &);
void initializeResolveAggregateArgumentsLPMPass(llvm::PassRegistry &);
void initializeResolveOCLAtomicsLPMPass(llvm::PassRegistry &);
void initializeResolveSampledImageBuiltinsLPMPass(llvm::PassRegistry &);
void initializeResourceAllocatorLPMPass(llvm::PassRegistry &);
void initializeRewriteLocalSizeLPMPass(llvm::PassRegistry &);
void initializeFPRoundingModeCoalescingPass(llvm::PassRegistry &);
void initializeRuntimeValueLegalizationPassPass(llvm::PassRegistry &);
void initializeRuntimeValueVectorExtractPassPass(llvm::PassRegistry &);
void initializeSampleCmpToDiscardPass(llvm::PassRegistry &);
void initializeSamplerPerfOptPassPass(llvm::PassRegistry &);
void initializeScalarArgAsPointerAnalysisLPMPass(llvm::PassRegistry &);
void initializeScalarizeFunctionLPMPass(llvm::PassRegistry &);
void initializeSimd32ProfitabilityAnalysisPass(llvm::PassRegistry &);
void initializeSetFastMathFlagsLPMPass(llvm::PassRegistry &);
void initializeSinkPointerConstAddPassPass(llvm::PassRegistry &);
void initializeSPIRMetaDataTranslationLPMPass(llvm::PassRegistry &);
void initializeSplitStructurePhisPassPass(llvm::PassRegistry &);
void initializeMergeMemFromBranchOptPass(llvm::PassRegistry &);
void initializeMergeScalarPhisPassPass(llvm::PassRegistry &);
void initializeSubgroup2DBlockIoResolutionLPMPass(llvm::PassRegistry &);
void initializeSpvPredicatedIOResolutionLPMPass(llvm::PassRegistry &);
void initializeSpvSubgroupMMAResolutionLPMPass(llvm::PassRegistry &);
void initializeSplitLoadsPass(llvm::PassRegistry &);
void initializeStatelessToStatefulPass(llvm::PassRegistry &);
void initializeStatelessOffsetNarrowingPass(llvm::PassRegistry &);
void initializeSubGroupFuncsResolutionLPMPass(llvm::PassRegistry &);
void initializeSubGroupReductionPatternPass(llvm::PassRegistry &);
void initializeTransformUnmaskedFunctionsPassLPMPass(llvm::PassRegistry &);
void initializeDisableInliningLPMPass(llvm::PassRegistry &);
void initializeDropTargetFunctionsLPMPass(llvm::PassRegistry &);
void initializeIndirectCallOptimizationPass(llvm::PassRegistry &);
void initializePromoteInt8TypePass(llvm::PassRegistry &);
void initializeDpasFuncsResolutionLPMPass(llvm::PassRegistry &);
void initializeDecompose2DBlockFuncsWithHoistingPass(llvm::PassRegistry &);
void initializeDecompose2DBlockFuncsPass(llvm::PassRegistry &);
void initializeLSCFuncsResolutionLPMPass(llvm::PassRegistry &);
void initializeConvertMSAAPayloadTo16BitPass(llvm::PassRegistry &);
void initializeInterfaceOptimizationPass(llvm::PassRegistry &);
void initializeMSAAInsertDiscardPass(llvm::PassRegistry &);
void initializeInlineRaytracingPass(llvm::PassRegistry &);
void initializeResolveOCLRaytracingBuiltinsLPMPass(llvm::PassRegistry &);
void initializeRayTracingIntrinsicLoweringPassPass(llvm::PassRegistry &);
void initializePromoteToScratchPassPass(llvm::PassRegistry &);
void initializePrepareLoadsStoresPassPass(llvm::PassRegistry &);
void initializeRayTracingPredicatedStackIDReleasePassPass(llvm::PassRegistry &);
void initializeShaderExitBehaviorPassPass(llvm::PassRegistry &);
void initializeEarlyRematPassPass(llvm::PassRegistry &);
void initializeLateRematPassPass(llvm::PassRegistry &);
void initializeRayInfoCSEPassPass(llvm::PassRegistry &);
void initializeLowerIntersectionAnyHitPass(llvm::PassRegistry &);
void initializeLSCCacheOptimizationPassPass(llvm::PassRegistry &);
void initializeBindlessKernelArgLoweringPassPass(llvm::PassRegistry &);
void initializeSplitPreparePassPass(llvm::PassRegistry &);
void initializeSplitAsyncPassPass(llvm::PassRegistry &);
void initializeMergeAllocasOCLPass(llvm::PassRegistry &);
void initializeMergeAllocasRaytracingPass(llvm::PassRegistry &);
void initializeAllocationBasedLivenessAnalysis(llvm::PassRegistry &);
void initializePrepareSyncRTHelpersPass(llvm::PassRegistry &);
void initializeSyncHandlingPassPass(llvm::PassRegistry &);
void initializeSyncDispatchRaysFinalizePass(llvm::PassRegistry &);
void initializeBindlessInlineDataPassPass(llvm::PassRegistry &);
void initializeRayTracingConstantCoalescingPassPass(llvm::PassRegistry &);
void initializeStackIDSchedulingPassPass(llvm::PassRegistry &);
void initializeRayTracingFinalizePassPass(llvm::PassRegistry &);
void initializePayloadSinkingPassPass(llvm::PassRegistry &);
void initializeDeadPayloadStoreEliminationPassPass(llvm::PassRegistry &);
void initializeInlineMergeCallsPassPass(llvm::PassRegistry &);
void initializeLowerGlobalRootSignaturePassPass(llvm::PassRegistry &);
void initializeRayTracingPrintfPostProcessPass(llvm::PassRegistry &);
void initializeRayTracingIntrinsicAnalysisLPMPass(llvm::PassRegistry &);
void initializeRayTracingIntrinsicResolutionLPMPass(llvm::PassRegistry &);
void initializeTraceRayInlinePrepPassPass(llvm::PassRegistry &);
void initializeTraceRayInlineLatencySchedulerPassPass(llvm::PassRegistry &);
void initializeTraceRayInlineLoweringPassPass(llvm::PassRegistry &);
void initializeRayTracingAddressSpaceAAWrapperPassPass(llvm::PassRegistry &);
void initializeRayTracingMemDSEPassPass(llvm::PassRegistry &);
void initializeOverrideTMaxPassPass(llvm::PassRegistry &);
void initializeAtomicPullSWWalkWrapperLoopImplementationPass(llvm::PassRegistry &);
void initializeTranslateToProgrammableOffsetsPassPass(llvm::PassRegistry &);
void initializeDynamicRayManagementPassPass(llvm::PassRegistry &);
void initializeVectorBitCastOptPass(llvm::PassRegistry &);
void initializeVectorPreProcessPass(llvm::PassRegistry &);
void initializeVectorProcessPass(llvm::PassRegistry &);
void initializeVerificationPassPass(llvm::PassRegistry &);
void initializeWaveAllJointReductionPass(llvm::PassRegistry &);
void initializeWGFuncResolutionLPMPass(llvm::PassRegistry &);
void initializeWIAnalysisPass(llvm::PassRegistry &);
void initializeWIFuncResolutionLPMPass(llvm::PassRegistry &);
void initializeWIFuncsAnalysisLPMPass(llvm::PassRegistry &);
void initializeWorkaroundAnalysisPass(llvm::PassRegistry &);
void initializeWAFMinFMaxPass(llvm::PassRegistry &);
void initializePingPongTexturesAnalysisPass(llvm::PassRegistry &);
void initializePingPongTexturesOptPass(llvm::PassRegistry &);
void initializeSampleMultiversioningPass(llvm::PassRegistry &);
void initializeLoopCanonicalizationPass(llvm::PassRegistry &);
void initializeLoopSplitWidePHIsPass(llvm::PassRegistry &);
void initializeLoopAllocaUpperboundPass(llvm::PassRegistry &);
void initializeLoopHoistConstantPass(llvm::PassRegistry &);
void initializeSpecialCasesDisableLICMPass(llvm::PassRegistry &);
void initializeMemOptPass(llvm::PassRegistry &);
void initializeLdStCombinePass(llvm::PassRegistry &);
void initializeBIFTransformsLPMPass(llvm::PassRegistry &);
void initializeThreadCombiningPass(llvm::PassRegistry &);
void initializeRegisterPressureEstimatePass(llvm::PassRegistry &);
void initializeLivenessAnalysisPass(llvm::PassRegistry &);
void initializeRegisterEstimatorPass(llvm::PassRegistry &);
void initializeVariableReuseAnalysisPass(llvm::PassRegistry &);
void initializeResourceLoopAnalysisPass(llvm::PassRegistry &);
void initializeTranslationTablePass(llvm::PassRegistry &);
void initializeTrivialUnnecessaryTGMFenceEliminationPass(llvm::PassRegistry &);
void initializeSLMConstPropPass(llvm::PassRegistry &);
void initializeBlendToDiscardPass(llvm::PassRegistry &);
void initializeCheckInstrTypesPass(llvm::PassRegistry &);
void initializeInstrStatisticPass(llvm::PassRegistry &);
void initializeHalfPromotionPass(llvm::PassRegistry &);
void initializeCapLoopIterationsPass(llvm::PassRegistry &);
void initializeFCmpPaternMatchPass(llvm::PassRegistry &);
void initializeCodeAssumptionLPMPass(llvm::PassRegistry &);
void initializeIGCInstructionCombiningPassPass(llvm::PassRegistry &);
void initializeIntDivConstantReductionPass(llvm::PassRegistry &);
void initializeIntDivRemIncrementReductionPass(llvm::PassRegistry &);
void initializeIntDivRemCombinePass(llvm::PassRegistry &);
void initializeGenRotatePass(llvm::PassRegistry &);
void initializeSynchronizationObjectCoalescingPass(llvm::PassRegistry &);
void initializeMoveStaticAllocasLPMPass(llvm::PassRegistry &);
void initializeNamedBarriersResolutionLPMPass(llvm::PassRegistry &);
void initializeUndefinedReferencesPassLPMPass(llvm::PassRegistry &);
void initializeUnreachableHandlingPass(llvm::PassRegistry &);
void initializeBreakdownIntrinsicPassLPMPass(llvm::PassRegistry &);
void initializeCatchAllLineNumberLPMPass(llvm::PassRegistry &);
void initializeDebugInfoPassPass(llvm::PassRegistry &);
void initializeRematChainsAnalysisPass(llvm::PassRegistry &);
void initializeVectorShuffleAnalysisPass(llvm::PassRegistry &);
void initializeIGCLivenessAnalysisPass(llvm::PassRegistry &);
void initializeIGCRegisterPressurePrinterPass(llvm::PassRegistry &);
void initializeIGCRegisterPressurePublisherPass(llvm::PassRegistry &);
void initializeIGCVectorizerPass(llvm::PassRegistry &);
void initializeIGCVectorCoalescerPass(llvm::PassRegistry &);
void initializeIGCFunctionExternalRegPressureAnalysisPass(llvm::PassRegistry &);
void initializePromoteConstantStructsPass(llvm::PassRegistry &);
void initializeLowerInvokeSIMDLPMPass(llvm::PassRegistry &);
void initializeRemoveCodeAssumptionsPass(llvm::PassRegistry &);
void initializePreprocessSPVIRLPMPass(llvm::PassRegistry &);
void initializeConvertUserSemanticDecoratorOnFunctionsLPMPass(llvm::PassRegistry &);
void initializePromoteSubByteLPMPass(llvm::PassRegistry &);
void initializeResolveConstExprCallsLPMPass(llvm::PassRegistry &);
void initializeHandleSpirvDecorationMetadataLPMPass(llvm::PassRegistry &);
void initializeNontemporalLoadsAndStoresInAssertLPMPass(llvm::PassRegistry &);
void initializeHandleDevicelibAssertLPMPass(llvm::PassRegistry &);
void initializeStackOverflowDetectionPassLPMPass(llvm::PassRegistry &);
void initializeRayTracingBiFModuleInitializePassPass(llvm::PassRegistry &);
void initializeRayTracingBiFModuleFinalizePassPass(llvm::PassRegistry &);
void initializeBufferBoundsCheckingLPMPass(llvm::PassRegistry &);
void initializeBufferBoundsCheckingPatcherLPMPass(llvm::PassRegistry &);
void initializeBfloatBuiltinsResolutionLPMPass(llvm::PassRegistry &);
void initializeBfloatFuncsResolutionLPMPass(llvm::PassRegistry &);
void initializeBarrierControlFlowOptimizationPass(llvm::PassRegistry &);
void initializeMinimumValidAddressCheckingLPMPass(llvm::PassRegistry &);
void initializeLoopCountAnalysisPass(llvm::PassRegistry &);
void initializeCollectLoopCountPass(llvm::PassRegistry &);
void initializeRemoveLoopDependencyPass(llvm::PassRegistry &);
void initializeResourceLoopUnrollPass(llvm::PassRegistry &);
void initializeInjectPrintfLPMPass(llvm::PassRegistry &);
void initializeWaveShuffleIndexSinkingPass(llvm::PassRegistry &);
void initializeInstructionHoistingOptimizationPass(llvm::PassRegistry &);
void initializeWaveBallotCSEPass(llvm::PassRegistry &);
void initializeProcessBICodeAssumptionLPMPass(llvm::PassRegistry &);
void initializeSimpleAluVectorizerPass(llvm::PassRegistry &);
void initializeRedundantOpsCSEPassPass(llvm::PassRegistry &);
void initializeStateIndexAddrChainCanonicalizePass(llvm::PassRegistry &);
