/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

/*
 * !!! DO NOT EDIT THIS FILE !!!
 *
 * This file was automagically crafted by GED's model parser.
 */


#ifndef GED_MAPPING_TABLES_H
#define GED_MAPPING_TABLES_H

#include "common/ged_compact_mapping_table.h"
#include "ged_ins_field.h"
extern const ged_compact_mapping_entry_t MappingTable0[4];
extern const ged_compact_mapping_entry_t MappingTable1[4];
extern const ged_compact_mapping_entry_t MappingTable2[4];
extern const ged_compact_mapping_entry_t MappingTable3[4];
extern const ged_compact_mapping_entry_t MappingTable4[4];
extern const ged_compact_mapping_entry_t MappingTable5[4];
extern const ged_compact_mapping_entry_t MappingTable6[4];
extern const ged_compact_mapping_entry_t MappingTable7[4];
extern const ged_compact_mapping_entry_t MappingTable8[16];
extern const ged_compact_mapping_entry_t MappingTable9[16];
extern const ged_compact_mapping_entry_t MappingTable10[16];
extern const ged_compact_mapping_entry_t MappingTable11[16];
extern const ged_compact_mapping_entry_t MappingTable12[16];
extern const ged_compact_mapping_entry_t MappingTable13[16];
extern const ged_compact_mapping_entry_t MappingTable14[2];
extern const ged_compact_mapping_entry_t MappingTable15[2];
extern const ged_compact_mapping_entry_t MappingTable16[2];
extern const ged_compact_mapping_entry_t MappingTable17[2];
extern const ged_compact_mapping_entry_t MappingTable18[16];
extern const ged_compact_mapping_entry_t MappingTable19[2];
extern const ged_compact_mapping_entry_t MappingTable20[2];
extern const ged_compact_mapping_entry_t MappingTable21[2];
extern const ged_compact_mapping_entry_t MappingTable22[2];
extern const ged_compact_mapping_entry_t MappingTable23[2];
extern const ged_compact_mapping_entry_t MappingTable24[2];
extern const ged_compact_mapping_entry_t MappingTable25[128];
extern const ged_compact_mapping_entry_t MappingTable26[128];
extern const ged_compact_mapping_entry_t MappingTable27[2];
extern const ged_compact_mapping_entry_t MappingTable28[2];
extern const ged_compact_mapping_entry_t MappingTable29[128];
extern const ged_compact_mapping_entry_t MappingTable30[128];
extern const ged_compact_mapping_entry_t MappingTable31[2];
extern const ged_compact_mapping_entry_t MappingTable32[2];
extern const ged_compact_mapping_entry_t MappingTable33[16];
extern const ged_compact_mapping_entry_t MappingTable34[2];
extern const ged_compact_mapping_entry_t MappingTable35[128];
extern const ged_compact_mapping_entry_t MappingTable36[128];
extern const ged_compact_mapping_entry_t MappingTable37[2];
extern const ged_compact_mapping_entry_t MappingTable38[2];
extern const ged_compact_mapping_entry_t MappingTable39[2];
extern const ged_compact_mapping_entry_t MappingTable40[2];
extern const ged_compact_mapping_entry_t MappingTable41[2];
extern const ged_compact_mapping_entry_t MappingTable42[2];
extern const ged_compact_mapping_entry_t MappingTable43[2];
extern const ged_compact_mapping_entry_t MappingTable44[2];
extern const ged_compact_mapping_entry_t MappingTable45[2];
extern const ged_compact_mapping_entry_t MappingTable46[2];
extern const ged_compact_mapping_entry_t MappingTable47[2];
extern const ged_compact_mapping_entry_t MappingTable48[2];
extern const ged_compact_mapping_entry_t MappingTable49[2];
extern const ged_compact_mapping_entry_t MappingTable50[2];
extern const ged_compact_mapping_entry_t MappingTable51[2];
extern const ged_compact_mapping_entry_t MappingTable52[2];
extern const ged_compact_mapping_entry_t MappingTable53[2];
extern const ged_compact_mapping_entry_t MappingTable54[2];
extern const ged_compact_mapping_entry_t MappingTable55[2];
extern const ged_compact_mapping_entry_t MappingTable56[2];
extern const ged_compact_mapping_entry_t MappingTable57[2];
extern const ged_compact_mapping_entry_t MappingTable58[2];
extern const ged_compact_mapping_entry_t MappingTable59[2];
extern const ged_compact_mapping_entry_t MappingTable60[2];
extern const ged_compact_mapping_entry_t MappingTable61[128];
extern const ged_compact_mapping_entry_t MappingTable62[128];
extern const ged_compact_mapping_entry_t MappingTable63[2];
extern const ged_compact_mapping_entry_t MappingTable64[2];
extern const ged_compact_mapping_entry_t MappingTable65[162];
extern const ged_compact_mapping_entry_t MappingTable66[162];
extern const ged_compact_mapping_entry_t MappingTable67[162];
extern const ged_compact_mapping_entry_t MappingTable68[162];
extern const ged_compact_mapping_entry_t MappingTable69[162];
extern const ged_compact_mapping_entry_t MappingTable70[162];
extern const ged_compact_mapping_entry_t MappingTable71[162];
extern const ged_compact_mapping_entry_t MappingTable72[162];
extern const ged_compact_mapping_entry_t MappingTable73[162];
extern const ged_compact_mapping_entry_t MappingTable74[162];
extern const ged_compact_mapping_entry_t MappingTable75[162];
extern const ged_compact_mapping_entry_t MappingTable76[162];
extern const ged_compact_mapping_entry_t MappingTable77[162];
extern const ged_compact_mapping_entry_t MappingTable78[162];
extern const ged_compact_mapping_entry_t MappingTable79[162];
extern const ged_compact_mapping_entry_t MappingTable80[162];
extern const ged_compact_mapping_entry_t MappingTable81[162];
extern const ged_compact_mapping_entry_t MappingTable82[162];
extern const ged_compact_mapping_entry_t MappingTable83[162];
extern const ged_compact_mapping_entry_t MappingTable84[162];
extern const ged_compact_mapping_entry_t MappingTable85[162];
extern const ged_compact_mapping_entry_t MappingTable86[162];
extern const ged_compact_mapping_entry_t MappingTable87[162];
extern const ged_compact_mapping_entry_t MappingTable88[162];
extern const ged_compact_mapping_entry_t MappingTable89[162];
extern const ged_compact_mapping_entry_t MappingTable90[162];
extern const ged_compact_mapping_entry_t MappingTable91[162];
extern const ged_compact_mapping_entry_t MappingTable92[162];
extern const ged_compact_mapping_entry_t MappingTable93[162];
extern const ged_compact_mapping_entry_t MappingTable94[162];
extern const ged_compact_mapping_entry_t MappingTable95[162];
extern const ged_compact_mapping_entry_t MappingTable96[162];
extern const ged_compact_mapping_entry_t MappingTable97[162];
extern const ged_compact_mapping_entry_t MappingTable98[162];
#endif // GED_MAPPING_TABLES_H
